import copy
from functools import singledispatchmethod
from typing import final

from dataspree.inference.core.config.configurable import Configurable

from dataspree.inference.core.signals import BaseSignal

import logging

logger = logging.getLogger(__name__)


class SignalHandler(Configurable):

    def __init__(self):
        Configurable.__init__(self)
        # Identifier used in order to check equality of objects derived from the instance that was deep or shallow
        # copied along the way
        self.__og_id: int = id(self)

    @final
    def handle_signal(self, signal: BaseSignal) -> None:
        self.process_signal(signal)
        logger.debug(f'Processing signal {str(signal)}.')
        if signal.append(self):
            self.propagate_signal(signal)

    def propagate_signal(self, signal: BaseSignal) -> None:
        pass

    def process_signal(self, signal: BaseSignal) -> None:
        self.signal_dispatcher(signal)

    def signal_dispatcher(self, signal: BaseSignal) -> None:
        """
        @dev implementing @singledispatchmethod.signal_dispatcher here, and decorating the implementations
        with @SignalHandler.signal_dispatcher.register is not possible.
        This causes the signal dispatcher to be not tied to the instance.
        If there are multiple classes that register a dispatch method, this would cause all of them to be active
        whenever one of them is supposed to be invoked.

        :param signal:  the signal that is supposed to be dispatched.
        """
        pass

    def __eq__(self, other) -> bool:
        return isinstance(other, SignalHandler) and self.__og_id == other.__og_id

    def __copy__(self):
        # TODO: Unit test this together with the equality constraint and a child class overriding __copy__
        #       themselves!
        cls = self.__class__
        shallow_copy = cls.__new__(cls)
        shallow_copy.__dict__.update(self.__dict__)

        shallow_copy.__og_id = self.__og_id
        return shallow_copy

    def __deepcopy__(self, memo):
        # TODO: Unit test this together with the equality constraint and a child class overriding __deepcopy__
        #       themselves!
        cls = self.__class__
        deep_copy = cls.__new__(cls)
        memo[id(self)] = deep_copy
        for k, v in self.__dict__.items():
            setattr(deep_copy, k, copy.deepcopy(v, memo))

        deep_copy.__og_id = self.__og_id
        return deep_copy
