from __future__ import annotations

import copy
from dataclasses import dataclass

from typing import List, TypeVar, Generic, Type
from typing import TYPE_CHECKING

from dataspree.inference.core.exceptions import ImplementationError

if TYPE_CHECKING:
    from dataspree.inference.core.node import Node

SourceT = TypeVar('SourceT')


@dataclass
class Route(Generic[SourceT]):
    _instance: SourceT

    @property
    def source_name(self):
        return self._instance.name

    @property
    def source_class(self):
        return type(self._instance)

    def contains(self, source):
        return self._instance == source


@dataclass
class BaseSignal(Generic[SourceT]):

    def __post_init__(self):
        self.routing: List[Route[SourceT]] = []

    def append(self, source: SourceT) -> bool:
        for route in self.routing:
            if route.contains(source):
                return False

        self.routing.append(Route(source))
        return True

    def __deepcopy__(self, memo):
        # Prevent a deepcopy from copying the routing list content

        # TODO: Unit test this together with the equality constraint and a child class overriding __deepcopy__
        #       themselves!
        cls = self.__class__
        deep_copy = cls.__new__(cls)
        memo[id(self)] = deep_copy

        items = dict(self.__dict__.items())
        del items['routing']
        for k, v in items:
            setattr(deep_copy, k, copy.deepcopy(v, memo))

        deep_copy.routing = copy.copy(self.routing)
        return deep_copy


@dataclass
class Signal(BaseSignal[SourceT]):
    _source: SourceT

    def __post_init__(self):
        BaseSignal.__post_init__(self)
        self.routing.append(Route(self._source))

    @property
    def source_class(self) -> Type[SourceT]:
        return type(self._source)

    def __deepcopy__(self, memo):
        # Prevent a deepcopy from copying the routing list content
        # Assert: source not copied; super routing content not copied

        # TODO: Unit test this together with the equality constraint and a child class overriding __deepcopy__
        #       themselves!
        cls = self.__class__
        deep_copy = cls.__new__(cls)
        memo[id(self)] = deep_copy
        items = dict(self.__dict__.items())
        del items['_source']
        del items['routing']

        for k, v in items:
            setattr(deep_copy, k, copy.deepcopy(v, memo))

        deep_copy.routing = copy.copy(self.routing)
        deep_copy._source = self._source
        return deep_copy


@dataclass
class NodeSignal(Signal['Node']):
    @property
    def source_name(self):
        return self._source.name

    def __str__(self) -> str:
        return (f'{type(self).__name__} from Node {self.source_name} via ' +
                '->'.join((f'{r.source_class.__name__}:{r.source_name}' for r in self.routing)))


@dataclass
class TerminatedNodeSignal(NodeSignal):
    pass
